@@ -1,45 +1,49 @@
 from __future__ import annotations
-import os
-import time
-import tempfile
 import logging
-from collections import namedtuple, UserDict
+import os
+import tempfile
+import time
+import types
+import typing
+from collections import UserDict
 from collections.abc import (
     Mapping as AbstractMapping,
     Sequence,
-    ValuesView,
     ItemsView as _ItemsView,
+    MappingView,
+    Set as AbstractSet,
     KeysView,
 )
-import typing
-from weakref import WeakKeyDictionary
-from typing import (
-    Optional,
-    Callable,
-    Type,
-    cast,
-    FrozenSet,
-    Mapping,
-    TYPE_CHECKING,
-    Tuple,
-    Any,
-    Union,
-    List,
-    Set,
-    Dict,
-    get_type_hints,
-)
-import types
-from itertools import chain
+from base64 import urlsafe_b64encode
 from enum import IntEnum
 from importlib import import_module
-
+from itertools import chain
+from typing import (
+    Any,
+    Callable,
+    cast,
+    Dict,
+    FrozenSet,
+    get_type_hints,
+    Iterable,
+    List,
+    Mapping,
+    Optional,
+    Set,
+    Tuple,
+    Type,
+    TYPE_CHECKING,
+    Union,
+    NamedTuple,
+)
+from weakref import WeakKeyDictionary
 
 import inflection
 from jinja2 import Environment, PackageLoader
 
 from .about import __version__
 from .typedef import parse_typedef, has_collect_class, ismetasubclass
+from .typing import T
 
 __version__  # Silence unused import warning.
 
@@ -69,6 +73,10 @@ class InvalidPostCoerceAttributeNames(ClassDefinitionError):
     ...
 
 
+class CoerceMappingValueError(ClassDefinitionError):
+    ...
+
+
 class ItemsView(_ItemsView):
     __slots__ = ()
     if TYPE_CHECKING:
@@ -78,6 +86,34 @@ class ItemsView(_ItemsView):
         yield from self._mapping
 
 
+class AtomicValuesView(MappingView):
+    __slots__ = ()
+
+    def __contains__(self, value):
+        for key, current_value in self._mapping:
+            if current_value == value:
+                return True
+        return False
+
+    def __iter__(self):
+        for key, value in self._mapping:
+            yield value
+
+
+class AtomicKeysView(MappingView, AbstractSet):
+    __slots__ = ()
+
+    @classmethod
+    def _from_iterable(self, it):
+        return set(it)
+
+    def __contains__(self, key):
+        return key in self._mapping
+
+    def __iter__(self):
+        yield from keys(self._mapping.__class__)
+
+
 def _convert_exception_to_json(item):
     assert isinstance(item, Exception), f"item {item!r} ({type(item)}) is not an exception!"
     if hasattr(item, "to_json"):
@@ -168,6 +204,13 @@ def make_fast_clear(fields, set_block, class_name):
     return code_template
 
 
+def make_fast_dumps(fields, class_name):
+    code_template = env.get_template("fast_dumps.jinja").render(
+        fields=fields, class_name=class_name
+    )
+    return code_template
+
+
 def make_fast_getset_item(
     fields: List[str],
     properties: List[str],
@@ -427,6 +470,7 @@ def create_proxy_property(
 
 CoerceMapping = Dict[str, Tuple[Union[Type, Tuple[Type, ...]], Callable]]
 EMPTY_FROZEN_SET = frozenset()
+EMPTY_MAPPING = FrozenMapping({})
 
 
 def __no_op_skip_get__(self):
@@ -437,6 +481,110 @@ def __no_op_skip_set__(self, value):
     return
 
 
+def keys(
+    instance_or_cls: Union[Type[T], T], *, all: bool = False
+) -> Union[AtomicKeysView, KeysView]:
+    if not isinstance(instance_or_cls, type):
+        cls = type(instance_or_cls)
+        instance = instance_or_cls
+    else:
+        cls = instance_or_cls
+        instance = None
+    if not isinstance(cls, Atomic):
+        raise TypeError(f"Can only call on Atomic-metaclassed types!, {cls}")
+    if instance is not None and not all:
+        return AtomicKeysView(instance)
+    if all:
+        return cls._all_accessible_fields
+    return cls._columns.keys()
+
+
+def values(instance) -> AtomicValuesView:
+    cls = type(instance)
+    if not isinstance(cls, Atomic):
+        raise TypeError("Can only call on Atomic-metaclassed types!")
+    if instance is not None:
+        return AtomicValuesView(instance)
+    raise TypeError(f"values of a {cls} object needs to be called on an instance of {cls}")
+
+
+def items(instance: T) -> ItemsView:
+    cls: Type[T] = type(instance)
+    if not isinstance(cls, Atomic):
+        raise TypeError("Can only call on Atomic-metaclassed types!")
+    if instance is not None:
+        return ItemsView(instance)
+    raise TypeError(f"items of a {cls} object needs to be called on an instance of {cls}")
+
+
+def get(instance: T, key, default=None) -> Optional[Any]:
+    cls: Type[T] = type(instance)
+    if not isinstance(cls, Atomic):
+        raise TypeError("Can only call on Atomic-metaclassed types!")
+    if instance is None:
+        raise TypeError(f"items of a {cls} object needs to be called on an instance of {cls}")
+    try:
+        return instance[key]
+    except KeyError:
+        return default
+
+
+def asdict(instance: T) -> Dict[str, Any]:
+    cls: Type[T] = type(instance)
+    if not isinstance(cls, Atomic):
+        raise TypeError("Must be an Atomic-metaclassed type!")
+    return instance._asdict()
+
+
+def astuple(instance: T) -> Tuple[Any, ...]:
+    cls: Type[T] = type(instance)
+    if not isinstance(cls, Atomic):
+        raise TypeError("Must be an Atomic-metaclassed type!")
+    return instance._astuple()
+
+
+def aslist(instance: T) -> List[Any]:
+    cls: Type[T] = type(instance)
+    if not isinstance(cls, Atomic):
+        raise TypeError("Must be an Atomic-metaclassed type!")
+    return instance._aslist()
+
+
+def unpack_coerce_mappings(mappings):
+    """
+    Transform:
+    {
+        "foo": (str, lambda val: item.encode("utf8")),
+        ("bar", "baz"): ((int, float), lambda val: f"{val}")
+    }
+    to
+    {
+        "foo": (str, lambda val: item.encode("utf8")),
+        "bar": ((int, float), lambda val: f"{val}"),
+        "baz": ((int, float), lambda val: f"{val}")
+    }
+    """
+    for key in mappings.keys():
+        if not isinstance(key, (str, tuple)):
+            raise CoerceMappingValueError(
+                f"Coerce mapping keys must either be a string or tuple of strings (Got {key!r})"
+            )
+        if isinstance(key, tuple) and not all(isinstance(prop_name, str) for prop_name in key):
+            raise CoerceMappingValueError(
+                f"Coerce mapping keys must either be a string or tuple of strings (Got {key!r})"
+            )
+        value = mappings[key]
+        if not isinstance(value, tuple) or len(value) != 2:
+            raise CoerceMappingValueError(
+                f"Coerce mapping key {key} must be Tuple[Union[Type, Tuple[Type, ...]], Callable[[T], U]], not {value!r}"
+            )
+        if isinstance(key, tuple):
+            for prop_name in key:
+                yield prop_name, value
+        else:
+            yield key, value
+
+
 class Atomic(type):
     __slots__ = ()
     REGISTRY = ReadOnly(set())
@@ -539,6 +687,7 @@ class Atomic(type):
                     f"{class_name} expects `__coerce__` to implement an AbstractMapping or None, "
                     f"not a {type(coerce_mappings)}"
                 )
+            coerce_mappings = dict(unpack_coerce_mappings(coerce_mappings))
             if not isinstance(support_cls_attrs["__coerce__"], ReadOnly):
                 support_cls_attrs["__coerce__"] = ReadOnly(coerce_mappings)
         coerce_mappings = cast(CoerceMapping, coerce_mappings)
@@ -786,7 +935,16 @@ class Atomic(type):
                 ns_globals,
                 ns_globals,
             )
-            class_cell_fixups.append(("clear", cast(types.FunctionType, ns_globals["clear"])))
+            exec(
+                compile(
+                    make_fast_dumps(combined_columns, class_name), "<make_fast_dumps>", mode="exec"
+                ),
+                ns_globals,
+                ns_globals,
+            )
+            class_cell_fixups.append(("_asdict", cast(types.FunctionType, ns_globals["_asdict"])))
+            class_cell_fixups.append(("_astuple", cast(types.FunctionType, ns_globals["_astuple"])))
+            class_cell_fixups.append(("_aslist", cast(types.FunctionType, ns_globals["_aslist"])))
             exec(
                 compile(
                     make_fast_getset_item(
@@ -845,6 +1003,9 @@ class Atomic(type):
             "clear",
             "__getitem__",
             "__setitem__",
+            "_asdict",
+            "_astuple",
+            "_aslist",
         ):
             # Move the autogenerated functions into the support class
             # Any overrides that *may* call them will be assigned
@@ -910,7 +1071,10 @@ class Atomic(type):
             tuple("_{}_".format(key) for key in combined_columns) + support_columns + extra_slots
         )
         dataclass_template = env.get_template("data_class.jinja").render(
-            class_name=class_name, slots=repr(dataclass_slots), data_class_attrs=data_class_attrs
+            class_name=class_name,
+            slots=repr(dataclass_slots),
+            data_class_attrs=data_class_attrs,
+            class_slots=current_class_slots,
         )
         ns_globals["_dataclass_attrs"] = data_class_attrs
         exec(compile(dataclass_template, "<dcs>", mode="exec"), ns_globals, ns_globals)
@@ -924,9 +1088,61 @@ class Atomic(type):
         klass.REGISTRY.add(support_cls)
         return support_cls
 
+    def from_json(cls: Type[T], data: Dict[str, Any]) -> T:
+        return cls(**data)
 
-Delta = namedtuple("Delta", ["state", "old", "new", "index"])
-LoggedDelta = namedtuple("LoggedDelta", ["timestamp", "key", "delta"])
+    def from_many_json(cls: Type[T], iterable: Iterable[Dict[str, Any]]) -> Tuple[T, ...]:
+        return tuple(cls(**item) for item in iterable)
+
+    def to_json(*instances):
+        """
+        Returns a dictionary compatible with json.dumps(...)
+        """
+        if isinstance(instances[0], type):
+            instances = instances[1:]
+        jsons = []
+        cached_class_binary_encoders = {}
+        types = {type(item) for item in instances}
+        cached_class_binary_encoders = {
+            instance_type: getattr(instance_type, "BINARY_JSON_ENCODERS", EMPTY_MAPPING)
+            for instance_type in types
+        }
+        for instance in instances:
+            instance_type = type(instance)
+            result = {}
+            special_binary_encoders = cached_class_binary_encoders[instance_type]
+            for key, value in instance:
+                # Support nested daos
+                if hasattr(value, "to_json"):
+                    value = value.to_json()
+                # Date/datetimes
+                elif hasattr(value, "isoformat"):
+                    value = value.isoformat()
+                elif isinstance(value, (bytearray, bytes)):
+                    if key in special_binary_encoders:
+                        value = special_binary_encoders[key](value)
+                    else:
+                        value = urlsafe_b64encode(value).decode()
+                elif not isinstance(value, (str,)) and isinstance(
+                    value, (AbstractMapping, Sequence)
+                ):
+                    value = _encode_simple_nested_base(value, immutable=True)
+                result[key] = value
+            jsons.append(result)
+        return tuple(jsons)
+
+
+class Delta(NamedTuple):
+    state: str
+    old: Any
+    new: Any
+    index: int
+
+
+class LoggedDelta(NamedTuple):
+    timestamp: float
+    key: str
+    delta: Delta
 
 
 class Undefined(object):
@@ -1027,6 +1243,78 @@ result._{{field}}_ = None
 """
 
 
+class ClassOrInstanceFuncsDescriptor:
+    __slots__ = "_class_function", "_instance_function", "_classes"
+
+    def __init__(self, class_function=None, instance_function=None):
+        self._classes = WeakKeyDictionary()
+        self._class_function = class_function
+        self._instance_function = instance_function
+
+    def instance_function(self, instance_function):
+        return type(self)(self._class_function, instance_function)
+
+    def class_function(self, class_function):
+        return type(self)(class_function, self._instance_function)
+
+    def __get__(self, instance, owner=None):
+        if instance is None:
+            try:
+                return self._classes[owner]
+            except KeyError:
+                value = self._classes[owner] = types.MethodType(self._class_function, owner)
+                return value
+        return types.MethodType(self._instance_function, instance)
+
+
+class IMapping(metaclass=Atomic):
+    """
+    Allow an instruct class instance to have the `keys()` function which is
+    mandatory to support **item unpacking.
+
+    This will collide with any property that's already named keys.
+    """
+
+    __slots__ = ()
+
+    @ClassOrInstanceFuncsDescriptor
+    def keys(cls, instance=None, *, all=False) -> Set[str]:
+        if instance is not None:
+            return keys(instance, all=all)
+        return keys(cls, all=all)
+
+    @keys.instance_function
+    def keys(self, *, all=False) -> Set[str]:
+        return keys(self, all=all)
+
+    @ClassOrInstanceFuncsDescriptor
+    def values(cls, item):
+        return values(item)
+
+    @values.instance_function
+    def values(self):
+        return values(self)
+
+    @ClassOrInstanceFuncsDescriptor
+    def items(cls, item):
+        return items(item)
+
+    @items.instance_function
+    def items(self):
+        return items(self)
+
+    @ClassOrInstanceFuncsDescriptor
+    def get(cls, instance, key, default=None):
+        return get(instance, key, default)
+
+    @get.instance_function
+    def get(self, key, default=None):
+        return get(self, key, default)
+
+
+Atomic.register_mixin("mapping", IMapping)
+
+
 def add_event_listener(*fields):
     def wrapper(func):
         func._event_listener_funcs = getattr(func, "_event_listener_funcs", ()) + fields
@@ -1072,9 +1360,9 @@ def _encode_simple_nested_base(iterable, *, immutable=None):
                 destination[key] = value
         return iterable
     elif isinstance(iterable, Sequence):
-        if immutable is None and isinstance(iterable, tuple):
+        if immutable is None and isinstance(iterable, (tuple, frozenset)):
             immutable = True
-        elif immutable is None and isinstance(iterable, list):
+        elif immutable is None and isinstance(iterable, (list, set)):
             immutable = False
         if immutable is None:
             try:
@@ -1093,36 +1381,55 @@ def _encode_simple_nested_base(iterable, *, immutable=None):
     return iterable
 
 
-class ClassOrInstanceFuncsDescriptor:
-    __slots__ = "_class_function", "_instance_function", "_classes"
+class JSONSerializable(metaclass=Atomic):
+    __slots__ = ()
 
-    def __init__(self, class_function=None, instance_function=None):
-        self._classes = WeakKeyDictionary()
-        self._class_function = class_function
-        self._instance_function = instance_function
+    def to_json(self) -> Dict[str, Any]:
+        return Atomic.to_json(self)[0]
 
-    def instance_function(self, instance_function):
-        return type(self)(self._class_function, instance_function)
+    def __json__(self) -> Dict[str, Any]:
+        return self.to_json()
 
-    def class_function(self, class_function):
-        return type(self)(class_function, self._instance_function)
+    @classmethod
+    def from_json(cls: Type[T], data: Dict[str, Any]) -> T:
+        return cls(**data)
 
-    def __get__(self, instance, owner=None):
-        if instance is None:
-            try:
-                return self._classes[owner]
-            except KeyError:
-                value = self._classes[owner] = types.MethodType(self._class_function, owner)
-                return value
-        return types.MethodType(self._instance_function, instance)
+    @classmethod
+    def from_many_json(cls: Type[T], iterable: Iterable[Dict[str, Any]]) -> Tuple[T, ...]:
+        return tuple(cls.from_json(item) for item in iterable)
 
 
-class Base(metaclass=Atomic):
+Atomic.register_mixin("json", JSONSerializable)
+
+
+class SimpleBase(metaclass=Atomic):
     __slots__ = ("_flags",)
     __setter_template__ = ReadOnly("self._{key}_ = val")
     __getter_template__ = ReadOnly("return self._{key}_")
     __defaults_init__ = ReadOnly(DEFAULTS)
 
+    def __new__(cls, *args, **kwargs):
+        # Get the edge class that has all the __slots__ defined
+        cls = cls._data_class
+        result = super().__new__(cls)
+        result._flags = Flags.UNCONSTRUCTED
+        assert "__dict__" not in dir(
+            result
+        ), "Violation - there should never be __dict__ on a slotted class"
+        return result
+
+    def __len__(self):
+        return len(keys(self.__class__))
+
+    def __contains__(self, item):
+        return item in self._all_accessible_fields
+
+    def __reduce__(self):
+        # Create an empty class then let __setstate__ in the autogen
+        # code to handle passing raw values.
+        data_class, support_cls, *_ = self.__class__.__mro__
+        return load_cls, (support_cls, (), {}), self.__getstate__()
+
     @classmethod
     def _create_invalid_type(cls, field_name, val, val_type, types_required):
         if len(types_required) > 1:
@@ -1144,78 +1451,6 @@ class Base(metaclass=Atomic):
     def _create_invalid_value(cls, message, *args, **kwargs):
         return ValueError(message, *args, **kwargs)
 
-    @ClassOrInstanceFuncsDescriptor
-    def keys(cls, all=False) -> Set[str]:
-        if all:
-            return cls._all_accessible_fields
-        return cls._columns.keys()
-
-    @keys.instance_function
-    def keys(self, all=False) -> Set[str]:
-        if all:
-            return self._all_accessible_fields
-        return self._columns.keys()
-
-    def __new__(cls, *args, **kwargs):
-        # Get the edge class that has all the __slots__ defined
-        cls = cls._data_class
-        result = super().__new__(cls)
-        result._flags = Flags.UNCONSTRUCTED
-        assert "__dict__" not in dir(
-            result
-        ), "Violation - there should never be __dict__ on a slotted class"
-        return result
-
-    def __len__(self):
-        return len(self.keys())
-
-    def __contains__(self, item):
-        return item in self.keys(all=True)
-
-    def get(self, key, default=None):
-        try:
-            return self[key]
-        except KeyError:
-            return default
-
-    @classmethod
-    def from_json(cls, data: dict):
-        return cls(**data)
-
-    def to_json(self) -> dict:
-        """
-        Returns a dictionary compatible with json.dumps(...)
-        """
-        result = {}
-        for key, value in self:
-            # Support nested daos
-            if hasattr(value, "to_json"):
-                value = value.to_json()
-            # Date/datetimes
-            elif hasattr(value, "isoformat"):
-                value = value.isoformat()
-            elif not isinstance(value, (str, bytearray, bytes)) and isinstance(
-                value, (AbstractMapping, Sequence)
-            ):
-                value = _encode_simple_nested_base(value, immutable=True)
-            result[key] = value
-        return result
-
-    def items(self):
-        return ItemsView(self)
-
-    def values(self):
-        return ValuesView(self)
-
-    def __json__(self):
-        return self.to_json()
-
-    def __reduce__(self):
-        # Create an empty class then let __setstate__ in the autogen
-        # code to handle passing raw values.
-        data_class, support_cls, *_ = self.__class__.__mro__
-        return load_cls, (support_cls, (), {}), self.__getstate__()
-
     def _handle_init_errors(self, errors, errored_keys, unrecognized_keys):
         if unrecognized_keys:
             fields = ", ".join(unrecognized_keys)
@@ -1239,18 +1474,18 @@ class Base(metaclass=Atomic):
         self._flags |= Flags.DEFAULTS_SET
         errors = []
         errored_keys = []
-        keys = self.keys()
-        if len(args) > len(keys):
+        class_keys = keys(self.__class__)
+        if len(args) > len(class_keys):
             raise TypeError(
-                f"__init__() takes {len(keys)} positional arguments but {len(args)} were given"
+                f"__init__() takes {len(class_keys)} positional arguments but {len(args)} were given"
             )
-        for key, value in zip(keys, args):
+        for key, value in zip(class_keys, args):
             try:
                 setattr(self, key, value)
             except Exception as e:
                 errors.append(e)
                 errored_keys.append(key)
-        class_keys = self.keys(all=True)
+        class_keys = self._all_accessible_fields
         for key in class_keys & kwargs.keys():
             value = kwargs[key]
             try:
@@ -1268,4 +1503,8 @@ class Base(metaclass=Atomic):
         pass
 
 
+class Base(SimpleBase, mapping=True, json=True):
+    __slots__ = ()
+
+
 AbstractMapping.register(Base)  # pytype: disable=attribute-error
